package lsedit;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JComponent;

public class RelationComponent extends JComponent implements MouseListener, MouseMotionListener   
{
	class DrawInfo extends Object {

		int m_style = 0;

		public void setStyle(int style)
		{
			m_style = style;
		} 

		public void		getBounds(Rectangle r)
		{
			r.setBounds(0,0,0,0);
		}
		
		public boolean contains(int x, int y)	// (x,y) w.r.t diagram
		{
			return false;
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
		}

		public boolean overEdgePoint(int x, int y)
		{
			return false;
		}

		public EdgePoint mouseOverEdgePoint(int x, int y)
		{
			return null;
		}
	}

	class ArcInfo extends DrawInfo {

		int	m_x;	// Top left hand corner of arc bounded area
		int m_y;

		/* arrow head at end of arc

             (x1, y1)
				|     \
				|       \
                |         \
				|           \
		(x1, y1+radius-1)-- (x1+radius-1, y1+radius-1)
		*/

		public ArcInfo()
		{
		}

		public void setArc(int x, int y)
		{
			m_x = x;
			m_y = y;
		}

		public void getBounds(Rectangle r)
		{
			r.setBounds(m_x, m_y, RECURSIVE_LOOP_DIAMETER, RECURSIVE_LOOP_DIAMETER);
		}

		public boolean contains(int x, int y)		// (x,y) w.r.t. diagram
		{
			int x1, y1;
			int	r;

			// Transform axis so that over centre of arc
		
			x1 = x - m_x - RECURSIVE_LOOP_RADIUS;
			y1 = y - m_y - RECURSIVE_LOOP_RADIUS;

			if (x1 >= 0 || y1 <= 0) {
				// We never paint inside the object the arc is attached to so anything in this quadrant is false
				r = (x1*x1 + y1*y1);

				// Are we within one pixel either way of lying on the loop that forms the arc
				if (r >= ((RECURSIVE_LOOP_RADIUS - 1) * (RECURSIVE_LOOP_RADIUS - NEAR_LOOP_THRESHOLD)) && r <= ((RECURSIVE_LOOP_RADIUS+1)*(RECURSIVE_LOOP_RADIUS+NEAR_LOOP_THRESHOLD))) {
					return(true);
			}	}
			return(false);
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
			if (m_style != Util.LINE_STYLE_DOTTED) {
				g.drawArc(0, 0, RECURSIVE_LOOP_DIAMETER, RECURSIVE_LOOP_DIAMETER, 270, 270);
			} else {
				int	i;

				for (i = 270; i < 540; i += RECURSIVE_LOOP_DOTTED_ANGLE+RECURSIVE_LOOP_DOTTED_SPACE) {
					g.drawArc(0, 0, RECURSIVE_LOOP_DIAMETER, RECURSIVE_LOOP_DIAMETER, i /* Start angle */, RECURSIVE_LOOP_DOTTED_ANGLE);
			}	}	 
			g.fillPolygon(m_arc_xp, m_arc_yp, 3);	// The arrow head
		}
	}	

	class ElidedInfo extends DrawInfo {

		public void getBounds(int srcX, int srcY, int dstX, int dstY, Rectangle r)
		{
			getLineBounds(srcX, srcY, dstX, dstY, r);
			includeArrowBounds(srcX, srcY, dstX, dstY, r);
		}

		public void paintComponent(Graphics g, int srcX, int srcY, int dstX, int dstY)
		{
			Util.drawArrow(g, getLs(), m_style, srcX, srcY, dstX, dstY, 1);
		}
	}

	class OutElidedInfo extends ElidedInfo {

		private EdgePoint m_src;
		int		m_dstX, m_dstY;

		public void setOutElidedInfo(EdgePoint srcPt, EdgePoint dstPt)
		{
			double srcX = srcPt.getX();
			double srcY = srcPt.getY();
			double dx   = srcX - dstPt.getX();
			double dy   = srcY - dstPt.getY();
			double df   = Math.sqrt(dx*dx + dy*dy);

			// Make line length twice arrow length

			m_dstX = (int) (srcX - dx*2*Util.ARROW_L/df);
			m_dstY = (int) (srcY - dy*2*Util.ARROW_L/df);
			m_src  = srcPt;
		}

		public void getBounds(Rectangle r)
		{
			getBounds((int) m_src.getX(), (int) m_src.getY(), m_dstX, m_dstY, r);
		}

		// Test to see if (x,y) is near line from (x1,y1) to (x2,y2)

		public boolean contains(int x, int y) // (x,y) w.r.t diagram
		{
			return(nearLine((int) m_src.getX(), (int) m_src.getY(), m_dstX, m_dstY, x, y));
		}

		public EdgePoint mouseOverEdgePoint(int x, int y)
		{
			if (nearPoint((int) m_src.getX(), (int) m_src.getY(), x, y)) {
				return(m_src);
			}
			return(null);
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
			paintComponent(g, (int) (m_src.getX() - shiftX), (int) (m_src.getY() - shiftY), m_dstX  - shiftX, m_dstY  - shiftY);
		}
	}

	class InElidedInfo extends ElidedInfo {

		int		m_srcX, m_srcY;
		private EdgePoint m_dst;

		public void setInElidedInfo(EdgePoint srcPt, EdgePoint dstPt)
		{
			double dstX = dstPt.getX();
			double dstY = dstPt.getY();
			double dx   = srcPt.getX() - dstX;
			double dy   = srcPt.getY() - dstY;
			double df   = Math.sqrt(dx*dx + dy*dy);
			   
			// Make line length twice arrow length

			m_srcX = (int) (dstPt.getX() + (dx*2*Util.ARROW_L/df));
			m_srcY = (int) (dstPt.getY() + (dy*2*Util.ARROW_L/df));
			m_dst  = dstPt;
		}

		public void getBounds(Rectangle r)
		{
			getBounds(m_srcX, m_srcY, (int) m_dst.getX(), (int) m_dst.getY(), r);
		}

		// Test to see if (x,y) is near line from (x1,y1) to (x2,y2)

		public boolean contains(int x, int y) // (x,y) w.r.t diagram
		{
			return(nearLine(m_srcX, m_srcY, (int) m_dst.getX(), (int) m_dst.getY(), x, y));
		}

		public EdgePoint mouseOverEdgePoint(int x, int y)
		{
			if (nearPoint((int) m_dst.getX(), (int) m_dst.getY(), x, y)) {
				return(m_dst);
			}
			return(null);
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
			paintComponent(g, m_srcX - shiftX, m_srcY - shiftY, (int) (m_dst.getX()  - shiftX), (int) (m_dst.getY()  - shiftY));
		}
	}

	class EdgeInfo extends DrawInfo {

		private int		  m_embellished;

		private EdgePoint m_src;
		private EdgePoint m_dst;

		public void setEdgeInfo(EdgePoint src, EdgePoint dst)
		{
/*
			if (src == null) {
				System.out.println("Source edgeInfo is null");
			}
 			if (dst == null) {
				System.out.println("Destination edgeInfo is null");
			}
 */
			m_src  = src;
			m_dst  = dst;
		}

		public void setEmbellished(int embellished)
		{
			m_embellished = embellished;
		}

		public void getBounds(Rectangle r)
		{
			getLineBounds((int) m_src.getX(), (int) m_src.getY(), (int) m_dst.getX(), (int) m_dst.getY(), r);

			if ((m_embellished & DRAW_TAIL_MARK) != 0) {
				includeTailBounds((int) m_src.getX(), (int) m_src.getY(), r);
			}
			if ((m_embellished & DRAW_CENTRE_MARK) != 0) {
				includeCentreBounds((int) m_src.getX(), (int) m_src.getY(), (int) m_dst.getX(), (int) m_dst.getY(), r);
			}
			if ((m_embellished & DRAW_ARROW_MARK) != 0) {
				includeArrowBounds((int) m_src.getX(), (int) m_src.getY(), (int) m_dst.getX(), (int) m_dst.getY(), r);
			}
		}

		// Test to see if (x,y) is near line from (x1,y1) to (x2,y2)

		public boolean contains(int x, int y) // (x,y) w.r.t diagram
		{
			return(nearLine((int) m_src.getX(), (int) m_src.getY(), (int) m_dst.getX(), (int) m_dst.getY(), x, y));
		}

		public EdgePoint mouseOverEdgePoint(int x, int y)
		{
			if (nearPoint((int) m_src.getX(), (int) m_src.getY(), x, y)) {
				return(m_src);
			}
			if (nearPoint((int) m_dst.getX(), (int) m_dst.getY(), x, y)) {
				return(m_dst);
			}
			return(null);
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
			int srcX = (int) (m_src.getX() - shiftX);
			int srcY = (int) (m_src.getY() - shiftY);
			int dstX = (int) (m_dst.getX() - shiftX);
			int dstY = (int) (m_dst.getY() - shiftY);

/*			if (m_src.getEntity().m_label.equals("code")) {
				System.out.println("EdgeInfo.paintComponent srcY=" + m_src.getY());
			}
*/
			
			if ((m_embellished & DRAW_ARROW_MARK) != 0) {
				Util.drawArrow(g, getLs(), m_style, srcX, srcY, dstX, dstY, arrowWeight());
			} else {
				Util.drawSegment(g, m_style, srcX, srcY, dstX, dstY);
			}

			if ((m_embellished & DRAW_TAIL_MARK) != 0) {
				g.fillOval(srcX-RelationInstance.NEAR_PIXEL_SIZE/2, srcY-RelationInstance.NEAR_PIXEL_SIZE/2, RelationInstance.NEAR_PIXEL_SIZE, RelationInstance.NEAR_PIXEL_SIZE);
			}
			if ((m_embellished & DRAW_CENTRE_MARK) != 0) {
				Color tc = g.getColor();
			
				g.setColor(HIGHLIGHT_COLOUR);
				g.fillRect((srcX + dstX - HIGHLIGHT_DIM)/2, (srcY + dstY - HIGHLIGHT_DIM)/2, HIGHLIGHT_DIM, HIGHLIGHT_DIM);
				g.setColor(tc);
			}
		}
	}

	// Constants

	protected final static Color  HIGHLIGHT_COLOUR       = Color.yellow;
	protected final static int	  HIGHLIGHT_DIM	         = 4;

	// Static methods must be in the top level class
	
	protected final static int	RECURSIVE_LOOP_RADIUS	 = 8;
	protected final static int	RECURSIVE_LOOP_DIAMETER  = RECURSIVE_LOOP_RADIUS * 2;
	protected final static int  RECURSIVE_LOOP_DOTTED_ANGLE = 15;
	protected final static int  RECURSIVE_LOOP_DOTTED_SPACE = 20;

	protected final static int[]  m_arc_xp = {0, 0, RECURSIVE_LOOP_RADIUS-1};
	protected final static int[]  m_arc_yp = {RECURSIVE_LOOP_RADIUS-1, 0, RECURSIVE_LOOP_RADIUS-1};
	
	protected final static int NEAR_POINT_THRESHOLD = 1;
	protected final static int NEAR_LINE_THRESHOLD2 = 1;		// Square of threshold to avoid using sqrt
	protected final static int NEAR_LOOP_THRESHOLD  = 1;

	public final static int DRAW_TAIL_MARK	 = 0x01;
	public final static int DRAW_CENTRE_MARK = 0x02;
	public final static int DRAW_ARROW_MARK  = 0x04;
		
	// Properties

	private RelationInstance	m_relationInstance;
	private Object				m_drawInfo;
	private int					timeStamp = -1;

	protected LandscapeEditorCore getLs()
	{
		return m_relationInstance.getLs();
	}

	// Stupidity: All these methods have to go in RelationComponent because java doesn't allow static inner functions

	public static void getLineBounds(int srcX, int srcY, int dstX, int dstY, Rectangle r)
	{
		int x1, y1, x2, y2, x, y;

		if (srcX <= dstX) {
			x1 = srcX;
			x2 = dstX;
		} else {
			x1 = dstX;
			x2 = srcX;
		}
		if (srcY <= dstY) {
			y1 = srcY;
			y2 = dstY;
		} else {
			y1 = dstY;
			y2 = srcY;
		}
		if (Util.LINE_W > 1) {
			int w, h, shift;

			w     = x2 - x1;
			h     = y2 - y1;
			shift = Util.LINE_W/2;

				
			if (h > 2*w) {
				x1 -= shift;
				x2 += shift;
			}
			if (w >= 2*h) {
				y1 -= shift;
				y2 += shift;
			}
		}
		r.setBounds(x1, y1, x2-x1, y2-y1);
	}
	
	public static void includeTailBounds(int srcX, int srcY, Rectangle r)
	{
		int x1, y1, x2, y2, x, y;

		x1 = r.x;
		y1 = r.y;
		x2 = x1 + r.width;
		y2 = y1 + r.height;

		x = srcX - RelationInstance.NEAR_PIXEL_SIZE/2;
		y = srcY - RelationInstance.NEAR_PIXEL_SIZE/2;

		if (x < x1) {
			x1 = x;
		}
		if (y < y1) {
			y1 = y;
		} 
		x += RelationInstance.NEAR_PIXEL_SIZE;
		y += RelationInstance.NEAR_PIXEL_SIZE;

		if (x > x2) {
			x2 = x;
		}
		if (y > y2) {
			y2 = y;
		}
		r.setBounds(x1, y1, x2-x1, y2-y1);
	}

	public static void includeCentreBounds(int srcX, int srcY, int dstX, int dstY, Rectangle r)
	{
		int x1, y1, x2, y2, x, y;

		x1 = r.x;
		y1 = r.y;
		x2 = x1 + r.width;
		y2 = y1 + r.height;

		x = (srcX + dstX - HIGHLIGHT_DIM)/2;
		y = (srcY + dstY - HIGHLIGHT_DIM)/2;

		if (x < x1) {
			x1 = x;
		}
		if (y < y1) {
			y1 = y;
		}
		x += HIGHLIGHT_DIM;
		y += HIGHLIGHT_DIM;
		if (x > x2) {
			x2 = x;
		}
		if (y > y2) {
			y2 = y;
		}
		r.setBounds(x1, y1, x2-x1, y2-y1);
	}

	public void includeArrowBounds(int srcX, int srcY, int dstX, int dstY, Rectangle r)
	{
		int x1, y1, x2, y2, x, y;

		x1 = r.x;
		y1 = r.y;
		x2 = x1 + r.width;
		y2 = y1 + r.height;

		int[] x3 = new int[3];
		int[] y3 = new int[3];

		if (Util.getArrow(srcX, srcY, dstX, dstY, x3, y3, m_relationInstance.getFrequency()) != 0) {
			int i;

			for (i = 0; i < 3; ++i) {
				x = x3[i];
				y = y3[i];

				if (x < x1) {
					x1 = x;
				}
				if (y < y1) {
					y1 = y;
				}
				if (x > x2) {
					x2 = x;
				}
				if (y > y2) {
					y2 = y;
		}	}	}
		r.setBounds(x1, y1, x2-x1, y2-y1);
	}
		
	// Test to see if (x,y) is near line from (x1,y1) to (x2,y2)

	public static boolean nearPoint(int srcX, int srcY, int x, int y) 
	{
		int diff;

		diff = x-srcX;
		if (diff < 0) {
			diff = -diff;
		}
		if (diff > NEAR_POINT_THRESHOLD) {
			return(false);
		}
		diff = y-srcY;
		if (diff < 0) {
			diff = -diff;
		}
		if (diff > NEAR_POINT_THRESHOLD) {
			return(false);
		}
		return(true);
	} 

	public static boolean nearLine(int srcX, int srcY, int dstX, int dstY, int x, int y) 
	{
		if (srcX == dstX && srcY == dstY) {
			// Degenerate case line becomes a point
			return(nearPoint(srcX, srcY, x, y));
		}
		
		// Determine distance to line and point R(x, y)
		//
		// Using parametric form for equation of line
		// L = P0 + tV	where V is vector from P0 to P1
		//
		// Find t such that vector R -> P0 + tV is orthogonal to V.
		// We then restrict to checking distances to
		// segment P0, P1 (0 < t < 1). 

		double rx = x;
		double ry = y;
		double vx = dstX - srcX;
		double vy = dstY - srcY; 
		
		double t = ((rx - srcX) * vx + (ry - srcY) * vy) / (vx*vx + vy*vy);

		if (t >= 0 && t <= 1.0) {
			// Only check for segment between P0 and P1
			double px = srcX + t * vx;
			double py = srcY + t * vy; 
			double lx = (rx - px);
			double ly = (ry - py);

			double dist2 = lx*lx + ly*ly;

			if (dist2 <= NEAR_LINE_THRESHOLD2) {
				return(true);
		}	}	
		return(false);
	}	
						
	protected Vector points() 
	{
		RelationInstance	relationInstance = m_relationInstance;
		Diagram				diagram			 = relationInstance.getDiagram();
		EntityInstance		root             = diagram.getDrawRoot();
		EntityInstance		drawSrc          = relationInstance.getDrawSrc();
		EntityInstance		drawDst          = relationInstance.getDrawDst();
		RelationClass		parentClass      = relationInstance.getRelationClass();
		OutElidedInfo		outElided = null;
		InElidedInfo		inElided  = null;
		EdgeInfo			edgeInfo;
		int					style;
		int					embellished;

		// We will travel from the source to the destination
		// going first up then down using as our points of traversal the inflection points

		EntityInstance endUp, endDown;

		boolean client   = !root.hasDescendantOrSelf(drawSrc);
		boolean supplier = !root.hasDescendantOrSelf(drawDst);

		if (!client && !supplier) {
			endUp   = drawSrc.commonAncestor(drawDst); // Common container
			endDown = endUp;
		} else if (client) {
			endUp   = drawSrc.getContainedBy();
			endDown = root.getContainedBy();
		} else {
			endUp   = root.getContainedBy();
			endDown = drawDst.getContainedBy();
		}

		if (endUp == null) {
			System.out.println("C Error: " + relationInstance + " " + client + " " + supplier);
			return null;
		}

		if (endDown == null) {
			System.out.println("S Error: " + relationInstance + " " + client + " " + supplier);
			return null;
		}

		// Not in the same container

		Vector v1 = new Vector();
		Vector v2 = new Vector();

		EntityInstance	e, sp, dp, ep, next;
	    EdgePoint		pt1, pt2, ppt;
		int				edgeMode = diagram.getEdgeMode();

		for (sp = drawSrc; (next = sp.getContainedBy()) != endUp;   sp = next);
		for (dp = drawDst; (next = dp.getContainedBy()) != endDown; dp = next);

		Rectangle spLyt = sp.getDiagramBounds();
		Rectangle dpLyt = dp.getDiagramBounds();

		if (relationInstance.getSrc() != drawSrc || relationInstance.getDst() != drawDst) {
			style = Util.LINE_STYLE_DOTTED;
		} else {
			style = relationInstance.getInheritedStyle();
		}

		embellished = DRAW_TAIL_MARK;
		if (relationInstance.isMarked(RelationInstance.GROUP_FLAG_MARK)) {
			embellished |= DRAW_CENTRE_MARK;
		}

		/* The following is a bit tricky.
		 *
		 * We build up two vectors of points which represent the
		 * the curve we want the edge to take. These sets are built
		 * one point at a time. 
		 *
		 * The first set builds a curve from the source entity 
		 * to the common container entity. The second set builds
		 * the curve from the destination to the common container.
		 * The second curve then has its points reversed an is 
		 * appended to the first curve. 
		 *
		 * Client/Supplier edge elision complicated things because
		 * it requires the ability to have discontinuous curves.
		 *
		 */

		ppt = pt1 = null;		// Previous point
		for (e = drawSrc; e != endUp; e = ep) {
			if (e == null) {
				MsgOut.println("Error in points: " + "(" + parentClass.getId() + " " +  relationInstance + ")");
				return null;
			}

			ep  = e.getContainedBy();
			pt1 = e.getOutPoint(relationInstance, edgeMode, spLyt, dpLyt);

			if (ppt != null) {
				if (e.isExitingRelationElided(parentClass)) {
					outElided = new OutElidedInfo();
					outElided.setOutElidedInfo(pt1, ppt);
					outElided.setStyle(style);
					v1.addElement(outElided);
					break;
				}
				edgeInfo = new EdgeInfo();
				edgeInfo.setEdgeInfo(ppt, pt1);
				edgeInfo.setStyle(style);
				edgeInfo.setEmbellished(embellished);
				embellished &= ~DRAW_TAIL_MARK;
				v1.addElement(edgeInfo);
			}
			ppt = pt1;
		}

		// We've found all points from source entity to 
		// the edge of its ancestor in the common container
	
		// Finally, find the points from the destination entity
		// to the edge of its ancestor in the common container.
		// We'll reverse the order later.

		embellished &= ~DRAW_TAIL_MARK;
		embellished |= DRAW_ARROW_MARK;

		ppt = pt2 = null;
		for (e = drawDst; e != endDown; e = ep) {
			ep  = e.getContainedBy();
			pt2 = e.getOutPoint(relationInstance, edgeMode, dpLyt, spLyt);

			if (ppt != null) {
				if (e.isEnteringRelationElided(parentClass)) {
					inElided = new InElidedInfo();
					inElided.setInElidedInfo(ppt, pt2);
					inElided.setStyle(style);
					v2.addElement(inElided);
					break;
				}
				edgeInfo = new EdgeInfo();
				edgeInfo.setEdgeInfo(pt2, ppt);
				edgeInfo.setStyle(style);
				edgeInfo.setEmbellished(embellished);
				embellished &= ~DRAW_ARROW_MARK;
				v2.addElement(edgeInfo);
			}
			ppt = pt2;
		} 

		if (outElided == null && inElided == null) {
			edgeInfo = new EdgeInfo();
			edgeInfo.setEdgeInfo(pt1, pt2);
			v1.addElement(edgeInfo);
		} else if (inElided == null) {
			v2.removeAllElements();
		} else if (outElided == null) {
			v1.removeAllElements();
		}

		for (int i = v2.size()-1; i >= 0; i--) {
			v1.addElement(v2.elementAt(i));
		}
		return v1;
	}

	// --------------
	// Public methods
	// --------------

	public RelationComponent(RelationInstance relationInstance)
	{
		setLayout(null);
		m_relationInstance = relationInstance;
		relationInstance.setSwingObject(this);

		addMouseListener(this);
		addMouseMotionListener(this);
	}

	protected int arrowWeight()
	{
		return ((LandscapeObject.g_infoShown == m_relationInstance || getLs().isWeightedArrowhead()) ? m_relationInstance.getFrequency() : 1);
	}

	public void computeBounds(Rectangle r)
	{
		DrawInfo drawInfo;

		if (m_drawInfo == null) {
			r.setBounds(0,0,0,0);
		} else if (m_drawInfo instanceof DrawInfo) {
			((DrawInfo) m_drawInfo).getBounds(r);
		} else {
			Enumeration en;
			int x1,y1,x2,y2, x, y;

			x1 = Integer.MAX_VALUE;
			x2 = Integer.MIN_VALUE;
			y1 = Integer.MAX_VALUE;
			y2 = Integer.MIN_VALUE;

			for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
				drawInfo = (DrawInfo) en.nextElement();
				drawInfo.getBounds(r);

				x = r.x;
				if (x < x1) {
					x1 = x;
				}
				y = r.y;
				if (y < y1) {
					y1 = y;
				}
				x += r.width;
				y += r.height;
				if (x > x2) {
					x2 = x;
				}
				if (y > y2) {
					y2 = y;
			}	}
			r.setBounds(x1,y1, x2-x1, y2-1);
	}	}

	public boolean contains(int x, int y)	// (x,y) w.r.t. my bounds
	{
//		System.out.println("RelationInstance.contains " + this);
		if (super.contains(x, y)) {
			// Translate so w.r.t diagram
			x += getX();
			y += getY();
			if (m_drawInfo != null) {
				if (m_drawInfo instanceof DrawInfo) {
					return ((DrawInfo) m_drawInfo).contains(x, y);
				}

				Enumeration en;

				for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
					if ( ((DrawInfo) en.nextElement()).contains(x, y)) {
						return(true);
		}	}	}	}
		return(false);
	}

	public EdgePoint mouseOverEdgePoint(int x, int y)
	{
		if (m_drawInfo != null) {
			if (m_drawInfo instanceof DrawInfo) {
				return ((DrawInfo) m_drawInfo).mouseOverEdgePoint(x, y);
			}

			Enumeration en;
			EdgePoint	ep;

			for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
				ep = ((DrawInfo) en.nextElement()).mouseOverEdgePoint(x,y);
				if (ep != null) {
					return(ep);
		}	}	}
		return(null);
	}

	// Compute the coordinates for the edge

	public void validate()
	{
//		System.out.println("RelationComponent.validate " + this);
/*		java.lang.Thread.dumpStack();
*/
		super.validate();

		RelationInstance	relationInstance = m_relationInstance;

		if (!relationInstance.isMarked(RelationInstance.VARIOUS_TYPE_MARKS)) {
			return;
		}

		EntityInstance		drawSrc, drawDst;
		Rectangle			r;
		Enumeration			en;
		int					style;

		drawSrc = relationInstance.getDrawSrc();

		if (drawSrc == null) {
			// We may validate a relation because we have changed its flags, not realising it is no longer in the diagram
			return;
		}

		drawDst = relationInstance.getDrawDst();
		if (drawDst == null) {
			// We may validate a relation because we have changed its flags, not realising it is no longer in the diagram
			return;
		}

		if (relationInstance.getSrc() != drawSrc || relationInstance.getDst() != drawDst) {
			style = Util.LINE_STYLE_DOTTED;
		} else {
			style = relationInstance.getInheritedStyle();
		}

		Diagram	  diagram   = relationInstance.getDiagram();
		int		  edgeMode  = diagram.getEdgeMode();

		if (relationInstance.isMarked(RelationInstance.LOOP_MARK)) {

			// This is a recursive loop from something to itself
			if (m_drawInfo == null || !(m_drawInfo instanceof ArcInfo)) {
				m_drawInfo = new ArcInfo();
			}
			((ArcInfo) m_drawInfo).setStyle(style);
//			System.out.println("RelationComponent:validate " + this + " recursive");
			((ArcInfo) m_drawInfo).setArc(drawSrc.getDiagramX() + drawSrc.getWidth() - RECURSIVE_LOOP_RADIUS, drawSrc.getDiagramY() - RECURSIVE_LOOP_RADIUS);
		} else if (!relationInstance.isMarked(RelationInstance.IN_OUT_ELIDED) && edgeMode != Do.DIRECT_EDGE_STATE && edgeMode != Do.SIDE_EDGE_STATE && drawSrc.getContainedBy() != drawDst.getContainedBy() && relationInstance.isMarked(RelationInstance.ELISION_MARK)) {
			// This is an ugly path of edges
			m_drawInfo = points();
//			System.out.println("RelationComponent:validate " + this + " normal" + m_points);
		} else {
			Rectangle srcLyt    = drawSrc.getDiagramBounds();
			Rectangle dstLyt    = drawDst.getDiagramBounds();
			EdgePoint pt1, pt2;

			pt1 = pt2 = null;
			if (drawSrc.hasDescendant(drawDst)) {
				if (edgeMode == Do.SIDE_EDGE_STATE) {
					pt1 = drawSrc.getLeftOutPoint(relationInstance);
					pt2 = drawDst.getLeftOutPoint(relationInstance);
				} else {
					pt1 = drawSrc.getTopOutPoint(relationInstance);
					pt2 = drawDst.getTopOutPoint(relationInstance);
				}
			} else if (drawDst.hasDescendant(drawSrc)) {
				if (edgeMode == Do.SIDE_EDGE_STATE) {
					pt1 = drawSrc.getRightOutPoint(relationInstance);
					pt2 = drawDst.getRightOutPoint(relationInstance);
				} else {
					pt1 = drawSrc.getBottomOutPoint(relationInstance);
					pt2 = drawDst.getBottomOutPoint(relationInstance);
				}
			} else {
				pt1 = drawSrc.getOutPoint(relationInstance, edgeMode, srcLyt /* drawSrc.diagramBounds */, dstLyt /* drawDst.diagramBounds */);
				pt2 = drawDst.getOutPoint(relationInstance, edgeMode, dstLyt /* drawDst.diagramBounds */, srcLyt /* drawSrc.diagramBounds */);
			}

			if (relationInstance.isMarked(RelationInstance.IN_OUT_ELIDED)) {
				OutElidedInfo out = null;
				InElidedInfo  in  = null;
				DrawInfo	  drawInfo;

				if (relationInstance.isMarked(RelationInstance.OUT_ELIDED_MARK)) {
					if (m_drawInfo != null) {
						if (m_drawInfo instanceof OutElidedInfo) {
							out = (OutElidedInfo) m_drawInfo;
						} else if (m_drawInfo instanceof Vector) {
							for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
								drawInfo = (DrawInfo) en.nextElement();
								if (drawInfo instanceof OutElidedInfo) {
									out = (OutElidedInfo) drawInfo;
									break;
					}	}	}	}
					if (out == null) {
						out = new OutElidedInfo();
					}
					// Show that the out edge is elided

					out.setOutElidedInfo(pt1, pt2);
					out.setStyle(relationInstance.getInheritedStyle());
//					System.out.println("RelationComponent:validate " + this + " out elided");
				} 

				if (relationInstance.isMarked(RelationInstance.IN_ELIDED_MARK)) {
					if (m_drawInfo != null) {
						if (m_drawInfo instanceof InElidedInfo) {
							in = (InElidedInfo) m_drawInfo;
						} else if (m_drawInfo instanceof Vector) {
							for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
								drawInfo = (DrawInfo) en.nextElement();
								if (drawInfo instanceof InElidedInfo) {
									in = (InElidedInfo) drawInfo;
									break;
					}	}	}	}
					if (in == null) {
						in = new InElidedInfo();
					}
					// Show that the in edge is elided
	   
					in.setInElidedInfo(pt1, pt2);
					in.setStyle(relationInstance.getInheritedStyle());
//					System.out.println("RelationComponent:validate " + this + " in elided");
				}

				if (out == null) {
					m_drawInfo = in;
				} else if (in == null) {
					m_drawInfo = out;
				} else {
					if (m_drawInfo instanceof Vector) {
						((Vector) m_drawInfo).removeAllElements();
					} else {
						m_drawInfo = new Vector();
					}
					((Vector) m_drawInfo).addElement(out);
					((Vector) m_drawInfo).addElement(in);
				}
			} else {
				int	embellished;

				// We will draw a direct edge between drawSrc and drawDst

				if (m_drawInfo == null || !(m_drawInfo instanceof EdgeInfo)) {
					m_drawInfo = new EdgeInfo();
				}

				EdgeInfo edgeInfo = (EdgeInfo) m_drawInfo;


				edgeInfo.setEdgeInfo(pt1, pt2);

/*
				if (drawSrc.getEntityLabel().equals("code")) {
					System.out.println("RelationInstance code source point " + pt1);
				}
 */
				edgeInfo.setStyle(style);
				embellished = DRAW_ARROW_MARK;
				if (relationInstance.isMarked(RelationInstance.GROUP_FLAG_MARK)) {
					embellished |= DRAW_CENTRE_MARK;
				}
				edgeInfo.setEmbellished(embellished);
		}	} 

		if (m_drawInfo != null) {
			r = new Rectangle();
			computeBounds(r);
			// We don't draw the right and bottom edge
			r.width  += 1;
			r.height += 1;
			setBounds(r);
//			System.out.println("RelationComponent.Bounds " + this + "=" + getBounds());
			repaint();
		}
	}

	// Paint the edge

	public void paintComponent(Graphics g)
	{
		int					shiftX, shiftY, x1, y1, x2, y2;

		shiftX = getX();
		shiftY = getY();
			
		g.setColor(m_relationInstance.getInheritedObjectColor());

		/*		For debugging */
		/*
			System.out.println("RelationComponent.paintComponent " + this + " {" + getX() + "," + getY() + "x" + getWidth() + "," + getHeight() + "}");
			g.drawRect(0, 0, getWidth()-1, getHeight()-1);

		if (m_src.getEntityLabel().equals("code")) {
			System.out.println("RelationComponent.paintComponent " + getBounds());
		}
		*/

		if (m_drawInfo != null) {

			if (m_drawInfo instanceof DrawInfo) {
				((DrawInfo) m_drawInfo).paintComponent(g, shiftX, shiftY);
			} else {
				Enumeration en;
				DrawInfo	drawInfo;

				for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
					drawInfo = (DrawInfo) en.nextElement();
					drawInfo.paintComponent(g, shiftX, shiftY);
	}	}	}	}	

	public String toString() 
	{
		return "RelationComponent: " + m_relationInstance;
	}

	// MouseListener interface

	public void mouseClicked(MouseEvent e)
	{
	}

	public void mouseEntered(MouseEvent e)
	{
//		System.out.println("RelationComponent.MouseEntered " + getBounds());
		m_relationInstance.mouseEntered();
	}

	public void mouseExited(MouseEvent e)
	{
//		System.out.println("RelationComponent.MouseExited");
		m_relationInstance.mouseExited();
	}

	public void mousePressed(MouseEvent ev)
	{
//		System.out.println("RelationComponent.mousePressed");
		m_relationInstance.mousePressed(ev, ev.getX() + getX(), ev.getY() + getY());
	}

	public void mouseReleased(MouseEvent ev)
	{
//		System.out.println("RelationInstance.mousePressed");
		m_relationInstance.mouseReleased(ev, ev.getX() + getX(), ev.getY() + getY());
	}

	// MouseMotionListener interface

	public void mouseDragged(MouseEvent ev)
	{
//		System.out.println("RelationComponent.mouseDragged");
		m_relationInstance.mouseDragged(ev, ev.getX() + getX(), ev.getY() + getY());
	}

	public void mouseMoved(MouseEvent ev)
	{
		m_relationInstance.mouseMoved(ev, ev.getX() + getX(), ev.getY() + getY());
	}
}
